/*-
 * Copyright (c) 2006 Norikatsu Shigemura
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <errno.h>

#define	SYM_COMPAT(sym, impl, ver)		\
	__sym_compat(sym, impl, ver)

#define	GLIBC_SYMBOL_2(sym, major, minor)	\
	GLIBC_##major##_##minor##_##sym

#define	GLIBC_SYMBOL_3(sym, major, minor, patch)	\
	GLIBC_##major##_##minor##_##patch##_##sym

#define	GLIBC_COMPAT_2(sym, major, minor)	\
	SYM_COMPAT(sym, GLIBC_SYMBOL_2(sym, major, minor), GLIBC_ ##major. ##minor);

#define	GLIBC_COMPAT_3(sym, major, minor, patch)	\
	SYM_COMPAT(sym, GLIBC_SYMBOL_3(sym, major, minor, patch), GLIBC_ ##major. ##minor. ##patch);



#define	DEBUG
#ifdef  DEBUG
#	include <stdio.h>
#	include <unistd.h>
#	define	dprintf(str, ...)	fprintf(stderr, "[%d] " str "\n", getpid() ,##__VA_ARGS__)
#	define	CLEAR_ERRNO		errno = 0
#else
#	define	dprintf(str, ...)	/* NOTHING TO DO */
#	define	CLEAR_ERRNO		/* NOTHING TO DO */
#endif

/* flags of open(2) */
#define LINUX_O_RDONLY			00
#define LINUX_O_WRONLY			01
#define LINUX_O_RDWR			02
#define LINUX_O_NDELAY			LINUX_O_NONBLOCK
#define LINUX_O_APPEND			02000
#define LINUX_O_SYNC			010000
#define LINUX_O_NONBLOCK		04000
#define LINUX_FASYNC			020000
#define LINUX_O_CREAT			0100
#define LINUX_O_TRUNC			01000
#define LINUX_O_EXCL			0200
#define LINUX_O_NOCTTY			0400

/* requests of ioctl(2) */
#if defined(__i386__) || defined(__amd64__)
#define _LINUX_IOC_NRBITS		8
#define _LINUX_IOC_TYPEBITS		8
#define _LINUX_IOC_SIZEBITS		14
#define _LINUX_IOC_DIRBITS		2
#else
#error	NOT SUPPORT ARCHITECTURE
#endif

#define _LINUX_IOC_NRMASK		( (1 << _LINUX_IOC_NRBITS  ) - 1 )
#define _LINUX_IOC_TYPEMASK		( (1 << _LINUX_IOC_TYPEBITS) - 1 )
#define _LINUX_IOC_SIZEMASK		( (1 << _LINUX_IOC_SIZEBITS) - 1 )
#define _LINUX_IOC_DIRMASK		( (1 << _LINUX_IOC_DIRBITS ) - 1 )

#define _LINUX_IOC_NRSHIFT		0
#define _LINUX_IOC_TYPESHIFT		(_LINUX_IOC_NRSHIFT   + _LINUX_IOC_NRBITS)
#define _LINUX_IOC_SIZESHIFT		(_LINUX_IOC_TYPESHIFT + _LINUX_IOC_TYPEBITS)
#define _LINUX_IOC_DIRSHIFT		(_LINUX_IOC_SIZESHIFT + _LINUX_IOC_SIZEBITS)

#define _LINUX_IOC_NONE			0U
#define _LINUX_IOC_WRITE		1U
#define _LINUX_IOC_READ			2U

#define _LINUX_IOC(dir, type, nr, size)		\
	(  ( (dir)  << _LINUX_IOC_DIRSHIFT  ) | \
	   ( (type) << _LINUX_IOC_TYPESHIFT ) | \
	   ( (nr)   << _LINUX_IOC_NRSHIFT   ) | \
	   ( (size) << _LINUX_IOC_SIZESHIFT )  )

#define _LINUX_IO(type, nr)			\
		_LINUX_IOC(_LINUX_IOC_NONE,		     (type), (nr), 0)
#define _LINUX_IOR(type, nr, size)		\
		_LINUX_IOC(_LINUX_IOC_READ,		     (type), (nr), sizeof(size))
#define _LINUX_IOW(type, nr, size)		\
		_LINUX_IOC(_LINUX_IOC_WRITE,		     (type), (nr), sizeof(size))
#define _LINUX_IOWR(type, nr, size)		\
		_LINUX_IOC(_LINUX_IOC_READ|_LINUX_IOC_WRITE, (type), (nr), sizeof(size))

/* sound device of ioctl(2) requests */
#define LINUX_SNDCTL_DSP_RESET                  _LINUX_IO  ('P',   0)
#define LINUX_SNDCTL_DSP_SYNC                   _LINUX_IO  ('P',   1)
#define LINUX_SNDCTL_DSP_SETFMT                 _LINUX_IOWR('P',   5, int)
#define LINUX_SNDCTL_DSP_STEREO                 _LINUX_IOWR('P',   3, int)
#define LINUX_SNDCTL_DSP_SPEED                  _LINUX_IOWR('P',   2, int)
#define LINUX_SNDCTL_DSP_CHANNELS               _LINUX_IOWR('P',   6, int)
#define LINUX_SNDCTL_DSP_SETFRAGMENT            _LINUX_IOWR('P',  10, int)
#define LINUX_SNDCTL_DSP_GETFMTS                _LINUX_IOR ('P',  11, int)
#define LINUX_SNDCTL_DSP_GETCAPS                _LINUX_IOR ('P',  15, int)
#define LINUX_SNDCTL_DSP_GETOPTR                _LINUX_IOR ('P',  18, linux_count_info)
#define LINUX_SNDCTL_DSP_GETOSPACE              _LINUX_IOR ('P',  12, linux_audio_buf_info)
#define LINUX_MIXER_READ(dev)                   _LINUX_IOR ('M', dev, int)
#define LINUX_SOUND_MIXER_DEVMASK               0xfe
#define LINUX_SOUND_MIXER_READ_DEVMASK          LINUX_MIXER_READ(LINUX_SOUND_MIXER_DEVMASK)
#define LINUX_SOUND_PCM_READ_CHANNELS           _LINUX_IOR ('P',   6, int)
#define LINUX_SOUND_PCM_WRITE_CHANNELS          LINUX_SNDCTL_DSP_CHANNELS

/* commands of fcntl(2) */
#define LINUX_F_DUPFD                   0
#define LINUX_F_GETFD                   1
#define LINUX_F_SETFD                   2
#define LINUX_F_GETFL                   3
#define LINUX_F_SETFL                   4
#define LINUX_F_GETLK                   5
#define LINUX_F_SETLK                   6
#define LINUX_F_SETLKW                  7
#define LINUX_F_SETOWN                  8
#define LINUX_F_GETOWN                  9

/* flags of shmget(2) */
#define	LINUX_SHM_R			0400
#define	LINUX_SHM_W			0200

/* flags of shmat(2) */
#define	LINUX_SHM_RDONLY		010000
#define	LINUX_SHM_RND			020000
#define	LINUX_SHM_REMAP			040000

/* commands of shmctl */
#define	LINUX_SHM_LOCK			11
#define	LINUX_SHM_UNLOCK		12

/* sizeof struct utsname */
#define	UTSNAME_LENGTH			65

/* errno */
#define	LINUX_EAGAIN			11


/* COMPAT: struct __sched_param */
struct linux__sched_param
{
	int				__sched_priority;
};

/* COMPAT: pthread_t
 * XXX: unsigned long int == struct pthread *
 */
typedef	unsigned long int		linux_pthread_t;

/* COMPAT: pthread_attr_t */
typedef	struct linux__pthread_attr_s
{
	int				__detachstate;
	int				__schedpolicy;
	struct linux__sched_param	__schedparam;
	int				__inheritsched;
	int				__scope;
	size_t				__guardsize;
	int				__stackaddr_set;
	void *				__stackaddr;
	size_t				__stacksize;
} linux_pthread_attr_t;

/* COMPAT: pthread_mutex */
typedef struct linux_pthread_mutex {
	int32_t dummy[6];
} linux_pthread_mutex_t;

/* COMPAT: pthread_mutexattr_t */
typedef struct linux_pthread_mutexattr {
	int32_t dummy;
} linux_pthread_mutexattr_t;


/* COMPAT: struct audio_buf_info */
#define linux_audio_buf_info                    audio_buf_info

/* COMPAT: struct count_info */
#define linux_count_info                        count_info

/* COMPAT: struct timespec */
struct linux_timespec {
        uint32_t                tv_sec;
        uint32_t                tv_nsec;
};

/* COMPAT: struct stat */
struct linux_stat {
        unsigned short int      st_dev;
        unsigned short int      __pad1;
        uint32_t                st_ino;
        unsigned short int      st_mode;
        unsigned short int      st_nlink;
        unsigned short int      st_uid;
        unsigned short int      st_gid;
        unsigned short int      st_rdev;
        unsigned short int      __pad2;
        uint32_t                st_size;
        uint32_t                st_blksize;
        uint32_t                st_blocks;
        struct linux_timespec   st_atimespec;
        struct linux_timespec   st_mtimespec;
        struct linux_timespec   st_ctimespec;
        uint32_t                __unused4;
        uint32_t                __unused5;
};

/*
typedef unsigned short int shmatt_t;
*/

#if 0
struct shmid_ds
  {
    struct ipc_perm shm_perm;           /* operation permission struct */
    int shm_segsz;                      /* size of segment in bytes */
    __time_t shm_atime;                 /* time of last shmat() */
    __time_t shm_dtime;                 /* time of last shmdt() */
    __time_t shm_ctime;                 /* time of last change by shmctl() */
    __pid_t shm_cpid;                   /* pid of creator */
    __pid_t shm_lpid;                   /* pid of last shmop */
    shmatt_t shm_nattch;                /* number of current attaches */
  };
#endif

/* COMPAT: struct utsname */
struct linux_utsname {
	char	sysname[UTSNAME_LENGTH];
	char	nodename[UTSNAME_LENGTH];
	char	release[UTSNAME_LENGTH];
	char	version[UTSNAME_LENGTH];
	char	machine[UTSNAME_LENGTH];
	char	domainname[UTSNAME_LENGTH];
};

